// Copyright 2017 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build goexperiment.darwincrypto && darwin && cgo

// Package darwin provides access to DarwinCrypto implementation functions.
// Check the variable Enabled to find out whether DarwinCrypto is available.
// If DarwinCrypto is not available, the functions in this package all panic.
package backend

import (
	"crypto"
	"crypto/cipher"
	"crypto/internal/boring/sig"
	"crypto/internal/fips140/nistec"
	"errors"
	"hash"
	_ "unsafe"

	"github.com/microsoft/go-crypto-darwin/xcrypto"
)

// Enabled controls whether FIPS crypto is enabled.
const Enabled = true

type BigInt = xcrypto.BigInt

func init() {
	sig.BoringCrypto()
}

const RandReader = xcrypto.RandReader

func SupportsHash(h crypto.Hash) bool {
	return xcrypto.SupportsHash(h)
}

func SupportsCurve(curve string) bool {
	switch curve {
	case "P-256", "P-384", "P-521":
		return true
	}
	return false
}

func SupportsRSAOAEPLabel(label []byte) bool {
	// CommonCrypto doesn't support labels
	// https://github.com/microsoft/go-crypto-darwin/issues/22
	return len(label) == 0
}

func SupportsPKCS1v15Hash(hash crypto.Hash) bool {
	switch hash {
	case crypto.SHA1, crypto.SHA224, crypto.SHA256, crypto.SHA384, crypto.SHA512, 0:
		return true
	}
	return false
}

func NewMD5() hash.Hash    { return xcrypto.NewMD5() }
func NewSHA1() hash.Hash   { return xcrypto.NewSHA1() }
func NewSHA224() hash.Hash { return xcrypto.NewSHA224() }
func NewSHA256() hash.Hash { return xcrypto.NewSHA256() }
func NewSHA384() hash.Hash { return xcrypto.NewSHA384() }
func NewSHA512() hash.Hash { return xcrypto.NewSHA512() }

func MD5(p []byte) (sum [16]byte)        { return xcrypto.MD5(p) }
func SHA1(p []byte) (sum [20]byte)       { return xcrypto.SHA1(p) }
func SHA224(p []byte) (sum [28]byte)     { return xcrypto.SHA224(p) }
func SHA256(p []byte) (sum [32]byte)     { return xcrypto.SHA256(p) }
func SHA384(p []byte) (sum [48]byte)     { return xcrypto.SHA384(p) }
func SHA512(p []byte) (sum [64]byte)     { return xcrypto.SHA512(p) }
func SHA512_224(p []byte) (sum [28]byte) { panic("cryptobackend: not available") }
func SHA512_256(p []byte) (sum [32]byte) { panic("cryptobackend: not available") }

func NewHMAC(h func() hash.Hash, key []byte) hash.Hash {
	return xcrypto.NewHMAC(h, key)
}

func NewAESCipher(key []byte) (cipher.Block, error) {
	return xcrypto.NewAESCipher(key)
}

func NewGCMTLS(c cipher.Block) (cipher.AEAD, error) {
	return xcrypto.NewGCMTLS(c)
}

func NewGCMTLS13(c cipher.Block) (cipher.AEAD, error) {
	return xcrypto.NewGCMTLS13(c)
}

type PublicKeyECDSA = xcrypto.PublicKeyECDSA
type PrivateKeyECDSA = xcrypto.PrivateKeyECDSA

func GenerateKeyECDSA(curve string) (X, Y, D xcrypto.BigInt, err error) {
	return xcrypto.GenerateKeyECDSA(curve)
}

func NewPrivateKeyECDSA(curve string, X, Y, D xcrypto.BigInt) (*xcrypto.PrivateKeyECDSA, error) {
	return xcrypto.NewPrivateKeyECDSA(curve, X, Y, D)
}

func NewPublicKeyECDSA(curve string, X, Y xcrypto.BigInt) (*xcrypto.PublicKeyECDSA, error) {
	return xcrypto.NewPublicKeyECDSA(curve, X, Y)
}

//go:linkname encodeSignature crypto/ecdsa.encodeSignature
func encodeSignature(r, s []byte) ([]byte, error)

//go:linkname parseSignature crypto/ecdsa.parseSignature
func parseSignature(sig []byte) (r, s []byte, err error)

func SignMarshalECDSA(priv *xcrypto.PrivateKeyECDSA, hash []byte) ([]byte, error) {
	return xcrypto.SignMarshalECDSA(priv, hash)
}

func VerifyECDSA(pub *xcrypto.PublicKeyECDSA, hash []byte, sig []byte) bool {
	return xcrypto.VerifyECDSA(pub, hash, sig)
}

func SupportsRSAKeyPrimes(primes int) bool {
	return primes == 2
}

func SupportsRSASaltLength(sign bool, salt int) bool {
	// CommonCrypto doesn't support custom salt length
	return salt == -1
}

type PublicKeyRSA = xcrypto.PublicKeyRSA
type PrivateKeyRSA = xcrypto.PrivateKeyRSA

func DecryptRSAOAEP(h, mgfHash hash.Hash, priv *xcrypto.PrivateKeyRSA, ciphertext, label []byte) ([]byte, error) {
	return xcrypto.DecryptRSAOAEP(h, priv, ciphertext, label)
}

func DecryptRSAPKCS1(priv *xcrypto.PrivateKeyRSA, ciphertext []byte) ([]byte, error) {
	return xcrypto.DecryptRSAPKCS1(priv, ciphertext)
}

func DecryptRSANoPadding(priv *xcrypto.PrivateKeyRSA, ciphertext []byte) ([]byte, error) {
	return xcrypto.DecryptRSANoPadding(priv, ciphertext)
}

func EncryptRSAOAEP(h, mgfHash hash.Hash, pub *xcrypto.PublicKeyRSA, msg, label []byte) ([]byte, error) {
	return xcrypto.EncryptRSAOAEP(h, pub, msg, label)
}

func EncryptRSAPKCS1(pub *xcrypto.PublicKeyRSA, msg []byte) ([]byte, error) {
	return xcrypto.EncryptRSAPKCS1(pub, msg)
}

func EncryptRSANoPadding(pub *xcrypto.PublicKeyRSA, msg []byte) ([]byte, error) {
	return xcrypto.EncryptRSANoPadding(pub, msg)
}

//go:linkname decodeKeyRSA crypto/rsa.decodeKey
func decodeKeyRSA(data []byte) (N, E, D, P, Q, Dp, Dq, Qinv xcrypto.BigInt, err error)

//go:linkname encodeKeyRSA crypto/rsa.encodeKey
func encodeKeyRSA(N, E, D, P, Q, Dp, Dq, Qinv xcrypto.BigInt) ([]byte, error)

//go:linkname encodePublicKeyRSA crypto/rsa.encodePublicKey
func encodePublicKeyRSA(N, E xcrypto.BigInt) ([]byte, error)

func GenerateKeyRSA(bits int) (N, E, D, P, Q, Dp, Dq, Qinv xcrypto.BigInt, err error) {
	data, err := xcrypto.GenerateKeyRSA(bits)
	if err != nil {
		return
	}
	return decodeKeyRSA(data)
}

func NewPrivateKeyRSA(N, E, D, P, Q, Dp, Dq, Qinv xcrypto.BigInt) (*xcrypto.PrivateKeyRSA, error) {
	encoded, err := encodeKeyRSA(N, E, D, P, Q, Dp, Dq, Qinv)
	if err != nil {
		return nil, err
	}
	return xcrypto.NewPrivateKeyRSA(encoded)
}

func NewPublicKeyRSA(N, E xcrypto.BigInt) (*xcrypto.PublicKeyRSA, error) {
	encoded, err := encodePublicKeyRSA(N, E)
	if err != nil {
		return nil, err
	}
	return xcrypto.NewPublicKeyRSA(encoded)
}

func SignRSAPKCS1v15(priv *xcrypto.PrivateKeyRSA, h crypto.Hash, hashed []byte) ([]byte, error) {
	return xcrypto.SignRSAPKCS1v15(priv, h, hashed)
}

func SignRSAPSS(priv *xcrypto.PrivateKeyRSA, h crypto.Hash, hashed []byte, saltLen int) ([]byte, error) {
	return xcrypto.SignRSAPSS(priv, h, hashed, saltLen)
}

func VerifyRSAPKCS1v15(pub *xcrypto.PublicKeyRSA, h crypto.Hash, hashed, sig []byte) error {
	return xcrypto.VerifyRSAPKCS1v15(pub, h, hashed, sig)
}

func VerifyRSAPSS(pub *xcrypto.PublicKeyRSA, h crypto.Hash, hashed, sig []byte, saltLen int) error {
	return xcrypto.VerifyRSAPSS(pub, h, hashed, sig, saltLen)
}

type PrivateKeyECDH = xcrypto.PrivateKeyECDH
type PublicKeyECDH = xcrypto.PublicKeyECDH

func ECDH(priv *xcrypto.PrivateKeyECDH, pub *xcrypto.PublicKeyECDH) ([]byte, error) {
	return xcrypto.ECDH(priv, pub)
}

func GenerateKeyECDH(curve string) (*xcrypto.PrivateKeyECDH, []byte, error) {
	return xcrypto.GenerateKeyECDH(curve)
}

func NewPrivateKeyECDH(curve string, bytes []byte) (*xcrypto.PrivateKeyECDH, error) {
	var key []byte
	switch curve {
	case "P-256":
		p, err := nistec.NewP256Point().ScalarBaseMult(bytes)
		if err != nil {
			return nil, err
		}
		key = p.Bytes()
	case "P-384":
		p, err := nistec.NewP384Point().ScalarBaseMult(bytes)
		if err != nil {
			return nil, err
		}
		key = p.Bytes()
	case "P-521":
		p, err := nistec.NewP521Point().ScalarBaseMult(bytes)
		if err != nil {
			return nil, err
		}
		key = p.Bytes()
	default:
		return nil, errors.New("NewPrivateKeyECDH: unsupported curve: " + curve)
	}
	return xcrypto.NewPrivateKeyECDH(curve, key, bytes)
}

func NewPublicKeyECDH(curve string, bytes []byte) (*xcrypto.PublicKeyECDH, error) {
	return xcrypto.NewPublicKeyECDH(curve, bytes)
}

func SupportsHKDF() bool {
	return true
}

func ExpandHKDF(h func() hash.Hash, pseudorandomKey, info []byte, keyLength int) ([]byte, error) {
	return xcrypto.ExpandHKDF(h, pseudorandomKey, info, keyLength)
}

func ExtractHKDF(h func() hash.Hash, secret, salt []byte) ([]byte, error) {
	return xcrypto.ExtractHKDF(h, secret, salt)
}

func SupportsPBKDF2() bool {
	return true
}

func PBKDF2(pass, salt []byte, iter, keyLen int, h func() hash.Hash) ([]byte, error) {
	return xcrypto.PBKDF2(pass, salt, iter, keyLen, h)
}

func SupportsTLS1PRF() bool {
	return false
}

func TLS1PRF(result, secret, label, seed []byte, h func() hash.Hash) error {
	panic("cryptobackend: not available")
}

func SupportsDESCipher() bool {
	return true
}

func SupportsTripleDESCipher() bool {
	return true
}

func NewDESCipher(key []byte) (cipher.Block, error) {
	return xcrypto.NewDESCipher(key)
}

func NewTripleDESCipher(key []byte) (cipher.Block, error) {
	return xcrypto.NewTripleDESCipher(key)
}

func SupportsRC4() bool { return true }

type RC4Cipher = xcrypto.RC4Cipher

func NewRC4Cipher(key []byte) (*RC4Cipher, error) { return xcrypto.NewRC4Cipher(key) }

func SupportsEd25519() bool {
	return true
}

type PublicKeyEd25519 = xcrypto.PublicKeyEd25519
type PrivateKeyEd25519 = xcrypto.PrivateKeyEd25519

func GenerateKeyEd25519() (PrivateKeyEd25519, error) {
	return xcrypto.GenerateKeyEd25519(), nil
}

func NewPrivateKeyEd25519(priv []byte) (PrivateKeyEd25519, error) {
	return xcrypto.NewPrivateKeyEd25519(priv)
}

func NewPublicKeyEd25519(pub []byte) (PublicKeyEd25519, error) {
	return xcrypto.NewPublicKeyEd25519(pub)
}

func NewPrivateKeyEd25519FromSeed(seed []byte) (PrivateKeyEd25519, error) {
	return xcrypto.NewPrivateKeyEd25519FromSeed(seed)
}

func SignEd25519(priv PrivateKeyEd25519, message []byte) ([]byte, error) {
	return xcrypto.SignEd25519(priv, message)
}

func VerifyEd25519(pub PublicKeyEd25519, message, sig []byte) error {
	return xcrypto.VerifyEd25519(pub, message, sig)
}

func SupportsDSA(l, n int) bool {
	return false
}

func GenerateParametersDSA(l, n int) (p, q, g xcrypto.BigInt, err error) {
	panic("cryptobackend: not available")
}

type PrivateKeyDSA struct{}
type PublicKeyDSA struct{}

func GenerateKeyDSA(p, q, g xcrypto.BigInt) (x, y xcrypto.BigInt, err error) {
	panic("cryptobackend: not available")
}

func NewPrivateKeyDSA(p, q, g, x, y xcrypto.BigInt) (*PrivateKeyDSA, error) {
	panic("cryptobackend: not available")
}

func NewPublicKeyDSA(p, q, g, y xcrypto.BigInt) (*PublicKeyDSA, error) {
	panic("cryptobackend: not available")
}

func SignDSA(priv *PrivateKeyDSA, hash []byte, parseSignature func([]byte) (xcrypto.BigInt, xcrypto.BigInt, error)) (r, s xcrypto.BigInt, err error) {
	panic("cryptobackend: not available")
}

func VerifyDSA(pub *PublicKeyDSA, hashed []byte, r, s xcrypto.BigInt, encodeSignature func(r, s xcrypto.BigInt) ([]byte, error)) bool {
	panic("cryptobackend: not available")
}
