// Copyright 2017 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build goexperiment.darwincrypto

package rsa

import (
	"crypto/internal/backend"
	"crypto/internal/backend/bbig"
	"errors"
	"math/big"
	_ "unsafe"

	"golang.org/x/crypto/cryptobyte"
	"golang.org/x/crypto/cryptobyte/asn1"
)

//go:linkname decodeKey
func decodeKey(data []byte) (N, E, D, P, Q, Dp, Dq, Qinv backend.BigInt, err error) {
	bad := func(e error) (N, E, D, P, Q, Dp, Dq, Qinv backend.BigInt, err error) {
		return nil, nil, nil, nil, nil, nil, nil, nil, e
	}
	input := cryptobyte.String(data)
	var version int
	n, e, d, p, q, dp, dq, qinv := new(big.Int), new(big.Int), new(big.Int), new(big.Int),
		new(big.Int), new(big.Int), new(big.Int), new(big.Int)
	// Parse the ASN.1 sequence
	if !input.ReadASN1(&input, asn1.SEQUENCE) {
		return bad(errors.New("invalid ASN.1 structure: not a sequence"))
	}
	if !input.ReadASN1Integer(&version) || version != 0 {
		return bad(errors.New("invalid ASN.1 structure: unsupported version"))
	}
	if !input.ReadASN1Integer(n) || !input.ReadASN1Integer(e) ||
		!input.ReadASN1Integer(d) || !input.ReadASN1Integer(p) ||
		!input.ReadASN1Integer(q) || !input.ReadASN1Integer(dp) ||
		!input.ReadASN1Integer(dq) || !input.ReadASN1Integer(qinv) {
		return bad(errors.New("invalid ASN.1 structure"))
	}
	return bbig.Enc(n), bbig.Enc(e), bbig.Enc(d), bbig.Enc(p), bbig.Enc(q),
		bbig.Enc(dp), bbig.Enc(dq), bbig.Enc(qinv), nil
}

//go:linkname encodeKey
func encodeKey(N, E, D, P, Q, Dp, Dq, Qinv backend.BigInt) ([]byte, error) {
	builder := cryptobyte.NewBuilder(nil)
	builder.AddASN1(asn1.SEQUENCE, func(b *cryptobyte.Builder) {
		b.AddASN1Int64(0)               // Add version as int64
		b.AddASN1BigInt(bbig.Dec(N))    // Add modulus
		b.AddASN1BigInt(bbig.Dec(E))    // Add public exponent
		b.AddASN1BigInt(bbig.Dec(D))    // Add private exponent
		b.AddASN1BigInt(bbig.Dec(P))    // Add prime1
		b.AddASN1BigInt(bbig.Dec(Q))    // Add prime2
		b.AddASN1BigInt(bbig.Dec(Dp))   // Add exponent1
		b.AddASN1BigInt(bbig.Dec(Dq))   // Add exponent2
		b.AddASN1BigInt(bbig.Dec(Qinv)) // Add coefficient
	})
	return builder.Bytes()
}

//go:linkname encodePublicKey
func encodePublicKey(N, E backend.BigInt) ([]byte, error) {
	builder := cryptobyte.NewBuilder(nil)
	builder.AddASN1(asn1.SEQUENCE, func(b *cryptobyte.Builder) {
		b.AddASN1BigInt(bbig.Dec(N)) // Add modulus
		b.AddASN1BigInt(bbig.Dec(E)) // Add public exponent
	})
	return builder.Bytes()
}
